//   Copyright (c)  2005,2006  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/RingTwinFloat.H"
#include "CoCoA/assert.H"
#include "CoCoA/ideal.H"
#include "CoCoA/FieldIdeal.H"
#include "CoCoA/MemPool.H"
#include "CoCoA/OpenMath.H"
#include "CoCoA/RingHom.H"

#include <iostream>
using std::ostream;
#include <memory>
using std::auto_ptr;
#include <vector>
using std::vector; // only in RingTwinFloatImpl::output
#include <algorithm>
using std::max;    // only in <anon>::ApproximatelyEqual  and in RingTwinFloatImpl::myCmp
using std::min;    // only in RingTwinFloatImpl::myIsEqualNZIgnoreSign
#include <limits>
using std::numeric_limits;
#include <cstdlib>
using std::abs;    // only in <anon>::ApproximatelyEqual
#include <cmath>
using std::sqrt;   // only in ctor for RingTwinFloatImpl


namespace CoCoA
{

  namespace // anonymous namespace for file local functions
  {

    typedef mpf_t* MultipleFloat_t;
    typedef mpf_t const* ConstMultipleFloat_t;

    inline MultipleFloat_t import(RingElemRawPtr rawx)
    {
      return static_cast<MultipleFloat_t>(rawx.myRawPtr());
    }

    inline ConstMultipleFloat_t import(RingElemConstRawPtr rawx)
    {
      return static_cast<ConstMultipleFloat_t>(rawx.myRawPtr());
    }


    // Returns precision (in bits) to which a and b are equal;
    // if a and b are equal to more than PrecBits then returns PrecBits
    size_t ApproximatelyEqual(const mpf_t a, const mpf_t b, size_t PrecBits)
    {
      CoCoA_ASSERT(mpf_sgn(a) != 0 && mpf_sgn(b) != 0);
//       {
//         std::cout<<"APPROX EQ: PrecBits="<<PrecBits<<std::endl;
//         std::cout<<"APPROX EQ: a=";
//       }
      if (mpf_sgn(a) == 0) return (mpf_sgn(b) == 0 ? PrecBits : 0);
      if (mpf_sgn(b) == 0) return 0;
///      if (mpf_sgn(a) != mpf_sgn(b)) return 0;
// DOES NOT WORK PROPERLY YET      if (mpf_eq(a, b, PrecBits)) return PrecBits;
      long int expa, expb;
      mpf_get_d_2exp(&expa, a); // just discard mantissa
      mpf_get_d_2exp(&expb, b); // just discard mantissa
//       std::cout<<"APPROX EQ exps are " << expa << " and " << expb << std::endl;
//       std::cout<<"APPROX EQ mants are " << manta << " and " << mantb << std::endl;
//       std::cout<<"APPROX EQ mpf_eq gives " << mpf_eq(a, b, PrecBits) << std::endl;
      if ((expa > expb && expa > expb+1) ||
          (expa < expb && expa+1 < expb))
        return 0;

      expa = max(expa, expb);
      // The rest is not terribly efficient, but do we get here that often???
      mpf_t diff;
      mpf_init2(diff, 8); // do not need much accuracy here
      if (mpf_sgn(a) == mpf_sgn(b))
        mpf_sub(diff, a, b);
      else // a and b have opposite signs.
        mpf_add(diff, a, b);
// {if(mpf_sgn(diff)==0)std::cout<<"APPROX EQ: diff is zero"<<std::endl;}
      if (mpf_sgn(diff) == 0) { mpf_clear(diff); return PrecBits; }
      long int expdiff;
      mpf_get_d_2exp(&expdiff, diff);
// {mantb=mpf_get_d_2exp(&expdiff, diff);std::cout<<"APPROX EQ diff mant="<<mantb<<" and exp="<<expb<<std::endl;}
      mpf_clear(diff);
      if (expdiff >= expa-1) return 0;
      return expa-expdiff-1; // result could be bigger than PrecBits
//       if (size_t(-expb) >= PrecBits) return PrecBits;
//       return size_t(-expb);
    }


    // Add to x a random number about the size of x/2^SNRbits.
    void perturb(mpf_t x, size_t SNRbits, size_t NoiseBits, gmp_randstate_t RandomSource)
    {
      long expx;
      mpf_get_d_2exp(&expx, x);
      mpf_t noise;
      mpf_init2(noise, NoiseBits);
      mpf_urandomb(noise, RandomSource, NoiseBits);
      // At this point noise is uniform in the interval [0,1]
      mpf_mul_ui(noise, noise, 2);
      mpf_sub_ui(noise, noise, 1);
      // Now noise is uniform in the interval [-1,1]
      long shift = expx - long(SNRbits);
      if (shift >= 0)
        mpf_mul_2exp(noise, noise, shift);
      else
        mpf_div_2exp(noise, noise, -shift);
      mpf_add(x, x, noise);
      mpf_clear(noise);
    }

  } // end of anonymous namespace


  /*-----------------------------------------------------------------*/
  /** \include RingTwinFloat.txt  */
  /*-----------------------------------------------------------------*/
  class RingTwinFloatImpl: public RingTwinFloatBase
  {
  private: // data members
    static const std::size_t myNumCompts = 2;
    mutable MemPool myMemMgr;           // MemPool must come before myZeroPtr, myOnePtr, and myMinusOnePtr
    std::auto_ptr<RingElem> myZeroPtr;  ///< Every ring stores its own zero.
    std::auto_ptr<RingElem> myOnePtr;   ///< Every ring stores its own one.
    std::auto_ptr<RingElem> myMinusOnePtr;
    const std::size_t myMantissaBits; ///< the precision requested (or 10, whichever is the greater)
    const std::size_t myGuardBits;    ///< compts must agree to at least myMantissaBits+myGuardBits
    const std::size_t mySNRBits;      ///< initially add noise mySNRBits "smaller" than mantissa
    const std::size_t myNoiseBits;    ///< initially the last myNoiseBits are "noise"
    const std::size_t myWorkingBits;  ///< the full precision actually used
    mutable gmp_randstate_t myRandomSource;

  private:
    RingTwinFloatImpl(size_t MantissaBits);
    RingTwinFloatImpl(const RingTwinFloatImpl&);           ///< NEVER DEFINED -- disable copy ctor
    RingTwinFloatImpl operator=(const RingTwinFloatImpl&); ///< NEVER DEFINED -- disable assignment
    ~RingTwinFloatImpl();
    friend RingTwinFloat NewRingTwinFloat(size_t MantissaBits); // the only function allowed to call the constructor

  public:
    virtual void myCharacteristic(ZZ& p) const;
    virtual bool IamCommutative() const;
    virtual bool IamIntegralDomain() const;
    virtual bool IamOrderedDomain() const;
    virtual bool IamField() const;
    virtual ConstRefRingElem myZero() const;
    virtual ConstRefRingElem myOne() const;
    virtual RingElemRawPtr myNew() const;
    virtual RingElemRawPtr myNew(long int n) const;
    virtual RingElemRawPtr myNew(const ZZ& n) const;
    virtual RingElemRawPtr myNew(ConstRawPtr rawt) const;
    virtual void myDelete(RawPtr rawx) const;                      // destroys x (incl all resources)
    virtual void mySwap(RawPtr rawx, RawPtr rawy) const;                             // swap(x, y)
    virtual void myAssign(RawPtr rawlhs, ConstRawPtr rawx) const;                  // lhs = x
    virtual void myAssign(RawPtr rawlhs, long n) const;                         // lhs = n
    virtual void myAssign(RawPtr rawlhs, const ZZ& N) const;                    // lhs = N
    virtual void myAssignZero(RawPtr rawlhs) const;                               // lhs = 0
    virtual void myNegate(RawPtr rawlhs, ConstRawPtr rawx) const;                  // lhs = -x
    virtual void myAdd(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const;        // lhs = x+y
    virtual void mySub(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const;        // lhs = x-y
    virtual void myMul(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const;        // lhs = x*y
    virtual void myDiv(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const;        // lhs = x/y
    virtual bool myIsDivisible(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const;// lhs = x/y, if divisible
    virtual bool myIsInvertible(ConstRawPtr rawx) const;                                // true iff x is invertible
    // No GCD virtual void myGcd(...) const;
    virtual void myPowerSmallExp(RawPtr rawlhs, ConstRawPtr rawx, unsigned long n) const;// lhs = x^n, n>1, x not -1,0,1
    virtual bool myIsPrintedWithMinus(ConstRawPtr rawx) const;
    virtual void myOutput(std::ostream& out, ConstRawPtr rawx) const;             // out << x
    virtual void myOutputSelf(std::ostream& out) const;                        // out << R
    virtual void myOutputSelf(OpenMathOutput& OMOut) const;                    // OMOut << R
    virtual void myOutput(OpenMathOutput& OMOut, ConstRawPtr rawx) const;         // OMOut << x
    virtual bool myIsZero(ConstRawPtr rawx) const;                                // x == 0
    virtual bool myIsOne(ConstRawPtr rawx) const;                                 // x == 1
    virtual bool myIsMinusOne(ConstRawPtr rawx) const;                            // x == -1
    virtual bool myIsInteger(ZZ& N, ConstRawPtr rawx) const;                      // true iff x is integer
    virtual bool myIsRational(ZZ& N, ZZ& D, ConstRawPtr rawx) const;              // true iff x is rational
    //    virtual bool myIsZeroAddMul: use default definition
    virtual bool myIsEqual(ConstRawPtr rawx, ConstRawPtr rawy) const;                // x == y
    virtual int myCmp(ConstRawPtr rawx, ConstRawPtr rawy) const;                     // result is <0, =0, >0 according as x<y, x=y, x>y
    virtual int mySign(ConstRawPtr rawx) const;                                   // -1,0,+1 according as x <0,=0,>0

    virtual ideal myIdealCtor(const std::vector<RingElem>& gens) const;

    virtual RingHom myCompose(const RingHom& phi, const RingHom& theta) const; // phi(theta(...))


    size_t myPrecisionBits() const;
  private:
    void myCheckConsistency(RawPtr rawx) const;
    size_t myMatchBitsForEqual(ConstRawPtr rawx) const;
    bool myIsEqualNZIgnoreSign(ConstRawPtr rawx, ConstRawPtr rawy) const; // abs(x) == abs(y), might throw InsufficientPrecision
  };




  RingTwinFloatImpl::RingTwinFloatImpl(size_t MantissaBits):
      myMemMgr(myNumCompts*sizeof(mpf_t), "RingTwinFloatImpl.myMemMgr"),
      myMantissaBits(MantissaBits<32?32:MantissaBits>8388608?8388608:MantissaBits),  // must be between 32 and 8388608
      myGuardBits(int(sqrt(double(myMantissaBits)))),
      mySNRBits(myMantissaBits+myGuardBits+myMantissaBits/2),
      myNoiseBits(17+myGuardBits),
      myWorkingBits(mySNRBits+myNoiseBits)
  {
    gmp_randinit_default(myRandomSource);
    myZeroPtr.reset(new RingElem(ring(this)));
    myOnePtr.reset(new RingElem(ring(this), 1));
    myMinusOnePtr.reset(new RingElem(ring(this), -1));
    myRefCountZero();
  }


  RingTwinFloatImpl::~RingTwinFloatImpl()
  {
    gmp_randclear(myRandomSource);
  }


  void RingTwinFloatImpl::myCharacteristic(ZZ& p) const
  {
    p = 0;
  }


  bool RingTwinFloatImpl::IamCommutative() const
  {
    return true;
  }


  bool RingTwinFloatImpl::IamIntegralDomain() const
  {
    return true;
  }


  bool RingTwinFloatImpl::IamOrderedDomain() const
  {
    return true;;
  }


  bool RingTwinFloatImpl::IamField() const
  {
    return true;
  }


  ConstRefRingElem RingTwinFloatImpl::myZero() const
  {
    return *myZeroPtr;
  }


  ConstRefRingElem RingTwinFloatImpl::myOne() const
  {
    return *myOnePtr;
  }


  RingElemRawPtr RingTwinFloatImpl::myNew() const
  {
    MultipleFloat_t ans = static_cast<MultipleFloat_t>(myMemMgr.alloc());
    for (size_t i=0; i < myNumCompts; ++i)
      mpf_init2(ans[i], myWorkingBits);
    return RingElemRawPtr(ans);
  }


  RingElemRawPtr RingTwinFloatImpl::myNew(long int n) const
  {
    if (n == 0) { return myNew(); }
    MultipleFloat_t ans = static_cast<MultipleFloat_t>(myMemMgr.alloc());
    for (size_t i=0; i < myNumCompts; ++i)
    {
      mpf_init2(ans[i], myWorkingBits);
      mpf_set_si(ans[i], n);
      if (i>0) perturb(ans[i], mySNRBits, myNoiseBits, myRandomSource);
    }
    return RingElemRawPtr(ans);
  }


  RingElemRawPtr RingTwinFloatImpl::myNew(const ZZ& N) const
  {
    if (IsZero(N)) { return myNew(); }
    MultipleFloat_t ans = static_cast<MultipleFloat_t>(myMemMgr.alloc());
    for (size_t i=0; i < myNumCompts; ++i)
    {
      mpf_init2(ans[i], myWorkingBits);
      mpf_set_z(ans[i], mpzref(N));
      if (i>0) perturb(ans[i], mySNRBits, myNoiseBits, myRandomSource);
    }
    return RingElemRawPtr(ans);
  }


  RingElemRawPtr RingTwinFloatImpl::myNew(ConstRawPtr rawcopy) const
  {
    MultipleFloat_t ans = static_cast<MultipleFloat_t>(myMemMgr.alloc());
    ConstMultipleFloat_t rhs(import(rawcopy));
    for (size_t i=0; i < myNumCompts; ++i)
    {
      // NB cannot use mpf_init_set here as it clobbers the precision
      mpf_init2(ans[i], myWorkingBits);
      mpf_set(ans[i], rhs[i]);
      // Should I perturb the copy???
    }
    return RingElemRawPtr(ans);
  }


  void RingTwinFloatImpl::myDelete(RawPtr rawx) const
  {
    MultipleFloat_t val(import(rawx));
    for (size_t i = myNumCompts; i-- > 0;)
      mpf_clear(val[i]);
    myMemMgr.free(rawx.myRawPtr());
  }


  void RingTwinFloatImpl::mySwap(RawPtr rawx, RawPtr rawy) const
  {
    MultipleFloat_t lhs(import(rawx));
    MultipleFloat_t rhs(import(rawy));
    for (size_t i=0; i < myNumCompts; ++i)
      mpf_swap(lhs[i], rhs[i]);
  }


  void RingTwinFloatImpl::myAssign(RawPtr rawlhs, ConstRawPtr rawx) const
  {
    MultipleFloat_t x(import(rawlhs));
    ConstMultipleFloat_t y(import(rawx));
    for (size_t i=0; i < myNumCompts; ++i)
      mpf_set(x[i], y[i]);
  }


  void RingTwinFloatImpl::myAssign(RawPtr rawlhs, long n) const
  {
    MultipleFloat_t x(import(rawlhs));
    for (size_t i=0; i < myNumCompts; ++i)
    {
      mpf_set_si(x[i], n);
      if (i>0) perturb(x[i], mySNRBits, myNoiseBits, myRandomSource);
    }
  }


  void RingTwinFloatImpl::myAssign(RawPtr rawlhs, const ZZ& N) const
  {
    MultipleFloat_t x(import(rawlhs));
    for (size_t i=0; i < myNumCompts; ++i)
    {
      mpf_set_z(x[i], mpzref(N));
      if (i>0) perturb(x[i], mySNRBits, myNoiseBits, myRandomSource);
    }
  }


  void RingTwinFloatImpl::myAssignZero(RawPtr rawlhs) const
  {
    MultipleFloat_t x(import(rawlhs));
    for (size_t i=0; i < myNumCompts; ++i)
      mpf_set_ui(x[i], 0);
  }


  void RingTwinFloatImpl::myNegate(RawPtr rawlhs, ConstRawPtr rawx) const
  {
    MultipleFloat_t x(import(rawlhs));
    ConstMultipleFloat_t y(import(rawx));
    for (size_t i=0; i < myNumCompts; ++i)
      mpf_neg(x[i], y[i]);
  }


  void RingTwinFloatImpl::myAdd(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    // Eliminate the trivial cases of x=0 or y=0; simplifies later code.
    if (myIsZero(rawx)) { myAssign(rawlhs, rawy); return; }
    if (myIsZero(rawy)) { myAssign(rawlhs, rawx); return; }

    ConstMultipleFloat_t b(import(rawx));
    ConstMultipleFloat_t c(import(rawy));
//     {
//       // Throw InsufficientPrecision if x = x+y  or  y = x+y
//       long expb, expc;
//       mpf_get_d_2exp(&expb, b[0]);
//       mpf_get_d_2exp(&expc, c[0]);
//       if (expb > expc && expb > expc+long(myMantissaBits+myGuardBits) ||
//           expb < expc && expb+long(myMantissaBits+myGuardBits) < expc)
//         throw RingTwinFloat::InsufficientPrecision();
//     }
    bool CancellationPossible = (mpf_sgn(b[0]) == -mpf_sgn(c[0]));
    for (size_t i=0; CancellationPossible && i < myNumCompts; ++i)
    {
      long expb, expc;
      mpf_get_d_2exp(&expb, b[i]);
      mpf_get_d_2exp(&expc, c[i]);
      // Line below could give harmless false positive if expb+1 or expc+1 causes overflow.
      CancellationPossible &= (expb == expc) || (expb == expc+1) || (expb+1 == expc);
      // CancellationPossible &= (abs(expb-expc) <= 1); // this is equivalent
    }

    if (CancellationPossible)
    {
      if (myIsEqualNZIgnoreSign(rawx, rawy)) // could throw InsufficientPrecision
      { myAssignZero(rawlhs); return; }
    }

    // Now we actually compute the sum.
    MultipleFloat_t a(import(rawlhs));
    for (size_t i=0; i < myNumCompts; ++i)
      mpf_add(a[i], b[i], c[i]);

    myCheckConsistency(rawlhs);
  }


  void RingTwinFloatImpl::mySub(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    // Eliminate the trivial cases of x=0 or y=0; simplifies later code.
    if (myIsZero(rawx)) { myNegate(rawlhs, rawy); return; }
    if (myIsZero(rawy)) { myAssign(rawlhs, rawx); return; }

    ConstMultipleFloat_t b(import(rawx));
    ConstMultipleFloat_t c(import(rawy));
//     {
//       // Throw InsufficientPrecision if x = x-y  or  -y = x-y
//       long expb, expc;
//       mpf_get_d_2exp(&expb, b[0]);
//       mpf_get_d_2exp(&expc, c[0]);
//       if (expb > expc && expb > expc+long(myMantissaBits+myGuardBits) ||
//           expb < expc && expb+long(myMantissaBits+myGuardBits) < expc)
//         throw RingTwinFloat::InsufficientPrecision();
//     }
    bool CancellationPossible = (mpf_sgn(b[0]) == mpf_sgn(c[0]));
    for (size_t i=0; CancellationPossible && i < myNumCompts; ++i)
    {
      long expb, expc;
      mpf_get_d_2exp(&expb, b[i]);
      mpf_get_d_2exp(&expc, c[i]);
      // Line below could give harmless false positive if expb+1 or expc+1 causes overflow.
      CancellationPossible &= (expb == expc) || (expb == expc+1) || (expb+1 == expc);
      // CancellationPossible &= (abs(expb-expc) <= 1); // this is equivalent
    }

    if (CancellationPossible)
    {
      if (myIsEqualNZIgnoreSign(rawx, rawy)) // could throw InsufficientPrecision
      { myAssignZero(rawlhs); return; }
    }

    // Now we actually compute the difference.
    MultipleFloat_t a(import(rawlhs));
    for (size_t i=0; i < myNumCompts; ++i)
      mpf_sub(a[i], b[i], c[i]);

    myCheckConsistency(rawlhs);
  }


  void RingTwinFloatImpl::myMul(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    if (myIsZero(rawx) || myIsZero(rawy)) { myAssignZero(rawlhs); return; }
//???     if (myIsOne(rawx)) { myAssign(lhs, y); return; }
//???     if (myIsOne(y)) { myAssign(lhs, x); return; }
    MultipleFloat_t a(import(rawlhs));
    ConstMultipleFloat_t b(import(rawx));
    ConstMultipleFloat_t c(import(rawy));
    for (size_t i=0; i < myNumCompts; ++i)
      mpf_mul(a[i], b[i], c[i]);

    myCheckConsistency(rawlhs);
  }


  void RingTwinFloatImpl::myDiv(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    CoCoA_ASSERT(!myIsZero(rawy));
    if (myIsZero(rawx)) { myAssignZero(rawlhs); return; }
//???    if (myIsOne(y)) { myAssign(rawlhs, rawx); return; }
    MultipleFloat_t a(import(rawlhs));
    ConstMultipleFloat_t b(import(rawx));
    ConstMultipleFloat_t c(import(rawy));
    for (size_t i=0; i < myNumCompts; ++i)
      mpf_div(a[i], b[i], c[i]);

    myCheckConsistency(rawlhs);
  }


  bool RingTwinFloatImpl::myIsDivisible(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    if (myIsZero(rawy)) return false;
    myDiv(rawlhs, rawx, rawy);
    return true;
  }


  bool RingTwinFloatImpl::myIsInvertible(ConstRawPtr rawx) const
  {
    return !myIsZero(rawx);
  }


  void RingTwinFloatImpl::myPowerSmallExp(RawPtr rawlhs, ConstRawPtr rawx, unsigned long n) const
  {
    // Assert that we have a genuinely non-trivial case.
    CoCoA_ASSERT(n > 1);
    CoCoA_ASSERT(!myIsZero(rawx) && !myIsOne(rawx) && !myIsMinusOne(rawx));
    myBinaryPower(rawlhs, rawx, n);
  }


  bool RingTwinFloatImpl::myIsPrintedWithMinus(ConstRawPtr rawx) const
  {
    return mySign(rawx) < 0;
  }


  void RingTwinFloatImpl::myOutput(ostream& out, ConstRawPtr rawx) const
  {
    // Special case if x happens to be an integer
    try
    {
      ZZ N;
      if (myIsInteger(N, rawx)) // could throw
      {
        out << N;
        return;
      }
    }
    catch (RingTwinFloat::InsufficientPrecision&) {} // discard InsufficientPrecision if it happens
    ConstMultipleFloat_t val(import(rawx));
    if (mpf_sgn(val[0]) == 0) { out << "0"; return; } // BUG??? Can this condition ever be true???

    const size_t SigFig = size_t(ceil(myMantissaBits*0.30103)); // 0.30103 approx log(2)/log(10)
    // Print format is 0.123456789*10^123
    // Print format is -0.123456789*10^(-123)
    // In the line below 10 = strlen("-0.") + strlen("*10^(-") + strlen(")")
    const size_t NumChars = 10+ SigFig + numeric_limits<long>::digits10;
    const int base = 10;
    std::vector<char> buffer(NumChars);
    mp_exp_t exp;
    const char* mantissa = mpf_get_str(&buffer[0], &exp, base, SigFig, val[0]);
    if (mpf_sgn(val[0]) < 0)
      out << "-0." << &mantissa[1];
    else
      out << "0." << mantissa;
    if (exp > 0) out << "*10^" << exp;
    if (exp < 0) out << "*10^(" << exp << ")";

//     for (size_t i=1; i < myNumCompts; ++i)
//     {
//       out << " && ";
//       mpf_get_str(&buffer[0], &exp, base, SigFig, val[i]);
//       if (mpf_sgn(val[i]) < 0) out << "-0." << &buffer[1];
//       else out << "0." << &buffer[0];
//       if (exp >= 0) out << "e+" << exp;
//       else out << "e" << exp;
//     }
  }


  void RingTwinFloatImpl::myOutputSelf(ostream& out) const
  {
    out << "RingTwinFloat(" << myMantissaBits << ")";
//??? myNumCompts???  myWorkingBits????
  }


  void RingTwinFloatImpl::myOutputSelf(OpenMathOutput& OMOut) const
  {
    OMOut->mySendApplyStart();
    OMOut << OpenMathSymbol("cocoa", "RingTwinFloat");
    OMOut << myMantissaBits;
//???    OMOut << myNumCompts;
//???    OMOut << myWorkingBits;
    OMOut->mySendApplyEnd();
  }


  void RingTwinFloatImpl::myOutput(OpenMathOutput& OMOut, ConstRawPtr rawx) const
  {
    ConstMultipleFloat_t val(import(rawx));
//?????    if (mpf_sgn(val[0]) == 0) return out << "0";

    const long SigFig = long(floor(myMantissaBits*0.3)); // 0.3 approx log(2)/log(10)
    // Print format is 0.123456789*10^123
    // Print format is -0.123456789*10^(-123)
    // In the line below 10 = strlen("-0.") + strlen("*10^(-") + strlen(")")
    const size_t NumChars = 10+ SigFig + numeric_limits<long>::digits10;
    const int base = 10;
    std::vector<char> buffer(NumChars);
    mp_exp_t exp;
    mpz_t mantissa;
    mpz_init(mantissa);
    mpf_get_str(&buffer[0], &exp, base, SigFig, val[0]);
    mpz_set_str(mantissa, &buffer[0], 10);
    OMOut->mySendApplyStart();
    OMOut << OpenMathSymbol("bigfloat1","bigfloat");
    OMOut << ZZ(mantissa)
          << long(base)
          << static_cast<long>(exp)-SigFig;  // BUG: slight risk of overflow
//     OMOut->OutputInteger(ZZ(mantissa));
//     OMOut->OutputInteger(10);
//     OMOut->OutputInteger(exp-SigFig);
    OMOut->mySendApplyEnd();

//     for (size_t i=1; i < myNumCompts; ++i)
//     {
//       out << " && ";
//       mpf_get_str(&buffer[0], &exp, base, SigFig, val[i]);
//       if (mpf_sgn(val[i]) < 0) out << "-0." << &buffer[1];
//       else out << "0." << &buffer[0];
//       if (exp >= 0) out << "e+" << exp;
//       else out << "e" << exp;
//     }
  }


  bool RingTwinFloatImpl::myIsZero(ConstRawPtr rawx) const
  {
    return mpf_sgn(import(rawx)[0]) == 0;
  }


  bool RingTwinFloatImpl::myIsOne(ConstRawPtr rawx) const
  {
    return myIsEqual(rawx, raw(*myOnePtr));
  }


  bool RingTwinFloatImpl::myIsMinusOne(ConstRawPtr rawx) const
  {
    return myIsEqual(rawx, raw(*myMinusOnePtr));
  }



  bool RingTwinFloatImpl::myIsInteger(ZZ& N, ConstRawPtr rawX) const
  {
    if (myIsZero(rawX)) { N = 0; return true; }


    ConstMultipleFloat_t x(import(rawX));

    long expx;
    mpf_get_d_2exp(&expx, x[0]);
    if (expx <= 0) return false;
    if (size_t(expx) > myMantissaBits) // cast to keep compiler quiet
      throw RingTwinFloat::InsufficientPrecision();
//    std::cout<<"NOT HUGE"<<std::endl;
    // x is non-zero and not huge and not in the open interval (-1,1).
    mpf_t tmp; mpf_init2(tmp, myMantissaBits+myGuardBits);
    mpf_set_ui(tmp, 1); mpf_div_ui(tmp, tmp, 2); // tmp is now 0.5
    mpf_add(tmp, tmp, x[0]);   // tmp is now x[0]+0.5
    mpf_floor(tmp, tmp);
    mpz_set_f(mpzref(N), tmp);
    mpf_clear(tmp);
    return myIsEqual(raw(RingElem(ring(this),N)), rawX); // Returns true/false/InsPrec as appropriate.
//    std::cout<<"GENERATED ZZ="<<N<<std::endl;
    mpf_sub(tmp, tmp, x[0]);
    long expdiff;
    double mant = mpf_get_d_2exp(&expdiff, tmp);
    mpf_clear(tmp);
//    std::cout<<"mant="<<mant<<std::endl;
    if (mant == 0) return true; // It was a perfect integer!
//    std::cout<<"eps has expdiff="<<expdiff<<"  and expx="<<expx<<std::endl;
    CoCoA_ASSERT(size_t(expx-expdiff)<=myMatchBitsForEqual(rawX) || myIsEqual(raw(RingElem(ring(this),N)),rawX));
    return size_t(expx-expdiff) > myMatchBitsForEqual(rawX); // NB myMatchBitsForEqual is unsigned!!

//     /// WARNING: This code was ripped off from myOutput...
//     /// WARNING: Its only merit is that it should work (but only slowly).
//     ConstMultipleFloat_t val(import(rawx));
//     const size_t SigFig = size_t(floor((myMantissaBits+myGuardBits)*0.30103)); // 0.30103 approx log(2)/log(10)
//     const size_t MaxDigits = size_t(floor(myMantissaBits*0.30103)); // max no of decimal digits allowed
//     const size_t NumChars = 2+ SigFig; // allow room for an initial minus sign and final NUL.
//     const int base = 10;
//     std::vector<char> buffer(NumChars);
//     mp_exp_t exp;
//     const char* mantissa = mpf_get_str(&buffer[0], &exp, base, SigFig, val[0]);
//     long LenMantissa = strlen(mantissa);
//     if (mpf_sgn(val[0]) < 0) --LenMantissa;

//     if (exp > 0 && size_t(exp) <= MaxDigits && exp >= LenMantissa)
//     {
//       mpz_set_str(mpzref(N), mantissa, 10);
//       for (int i=0; i < exp-LenMantissa; ++i)
//         N *= 10;
//       return true;
//     }
//     return false;

//     mpf_round_z(raw(n), import(rawx)[0]);
//     RingElem tmp(ring(this), 1);
// //     tmp /= 2;
// //     myAdd(raw(tmp), raw(tmp), x); // x + 1/2
// //     mpf_floor(import(raw(tmp))[0], import(raw(tmp))[0]);
// //     mpf_get_z(raw(n), import(raw(tmp))[0]);
//     tmp = n;
//     return myIsEqual(x, raw(tmp));
  }


  bool RingTwinFloatImpl::myIsRational(ZZ& N, ZZ& D, ConstRawPtr rawx) const
  {
    // The line below ensures that myIsInteger and myIsRational are consistent.
    if (myIsInteger(N, rawx)) { D = 1; return true; } // this might throw InsPrec

    // Now we know that x is not an integer (in ptic x is non-zero), and not huge.
    const size_t MatchBits = myMatchBitsForEqual(rawx);
    const unsigned long prec = MatchBits + 32; // extra 32 bits to absorb rounding error

    // Store the sign of x; the main algm work with positive values, will restore sign at the end.
    int sign = mpf_sgn(import(rawx)[0]);
    // Complain if x is tiny, i.e. if 1/x is huge.
    long expx;
    mpf_get_d_2exp(&expx, import(rawx)[0]);
    if (expx < 0 && size_t(-expx) >= myMantissaBits) throw RingTwinFloat::InsufficientPrecision();

    // This is main part; put in hi and lo upper and lower bounds for rational value.
    mpf_t lo, hi, tmp; // tmp gets used for various temporary results.
    mpf_init2(lo, prec); mpf_init2(hi, prec); mpf_init2(tmp, prec);
    mpf_set_ui(tmp, 1); mpf_div_2exp(tmp, tmp, long(MatchBits)-expx);
    mpf_abs(lo, import(rawx)[0]); mpf_sub(lo, lo, tmp);
    mpf_abs(hi, import(rawx)[0]); mpf_add(hi, hi, tmp);
//       std::cout<<"flt_eps=";mpf_out_str(stdout,10,0,tmp);std::cout<<std::endl;
//       std::cout<<"flt_lo=";mpf_out_str(stdout,10,0,lo);std::cout<<std::endl;
//       std::cout<<"flt_hi=";mpf_out_str(stdout,10,0,hi);std::cout<<std::endl;

    mpz_t N0, N1, D0, D1;
    mpz_init_set_ui(N0, 1);    mpz_init_set_ui(N1, 0);    mpz_init_set_ui(D0, 0);    mpz_init_set_ui(D1, 1);
    mpz_t int_lo, int_hi, ztmp;
    mpz_init(int_lo); mpz_init(int_hi); mpz_init(ztmp);

    mpz_set_f(int_lo, lo);
    mpz_set_f(int_hi, hi);
    while (mpz_cmp(int_lo, int_hi) == 0)
    {
//       std::cout<<"quot_lo=";mpz_out_str(stdout,10,int_lo);std::cout<<std::endl;
//       std::cout<<"quot_hi=";mpz_out_str(stdout,10,int_hi);std::cout<<std::endl;
//       std::cout<<"flt_lo=";mpf_out_str(stdout,10,0,lo);std::cout<<std::endl;
//       std::cout<<"flt_hi=";mpf_out_str(stdout,10,0,hi);std::cout<<std::endl;
      mpz_mul(ztmp, int_lo, N0);
      mpz_add(N1, N1, ztmp);
      mpz_mul(ztmp, int_lo, D0);
      mpz_add(D1, D1, ztmp);
      mpz_swap(N0, N1);
      mpz_swap(D0, D1);
      if (mpf_integer_p(lo)) break;
      mpf_set_z(tmp, int_lo);
      mpf_sub(lo, lo, tmp);
      mpf_ui_div(lo, 1, lo);
      mpf_sub(hi, hi, tmp);
      mpf_ui_div(hi, 1, hi);
      mpf_swap(lo, hi);
      mpz_set_f(int_lo, lo);
      mpz_set_f(int_hi, hi);
    }
    if (mpz_cmp(int_lo, int_hi) != 0)
    {
      mpf_ceil(lo, lo);
      mpz_set_f(int_lo, lo);
      mpz_mul(ztmp, int_lo, N0);
      mpz_add(N1, N1, ztmp);
      mpz_mul(ztmp, int_lo, D0);
      mpz_add(D1, D1, ztmp);
      mpz_swap(N0, N1);
      mpz_swap(D0, D1);
    }

    // Check that rational is "simple" compared to precision used.
    mpz_mul(ztmp, N0, D0);
    bool AnswerIsGood = true;
    if (mpz_sizeinbase(ztmp, 2) >= myMantissaBits) AnswerIsGood = false;
    if (sign < 0) mpz_neg(N0, N0); // make negative if input was negative
    mpz_swap(mpzref(N), N0);
    mpz_swap(mpzref(D), D0);
    mpz_clear(ztmp); mpz_clear(int_hi); mpz_clear(int_lo);
    mpz_clear(D1);      mpz_clear(D0);      mpz_clear(N1);      mpz_clear(N0);
    mpf_clear(tmp); mpf_clear(hi), mpf_clear(lo);
    if (!AnswerIsGood) throw RingTwinFloat::InsufficientPrecision();
    return true;
  }


  bool RingTwinFloatImpl::myIsEqual(ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    return myCmp(rawx, rawy) == 0; // myCmp may throw InsufficientPrecision
  }


  int RingTwinFloatImpl::myCmp(ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    ConstMultipleFloat_t a(import(rawx));
    ConstMultipleFloat_t b(import(rawy));
    // Deal with obvious cases based just on sign information.
    const int signx = mpf_sgn(a[0]);
    const int signy = mpf_sgn(b[0]);
    if (signy == 0) return signx;
    if (signx == 0) return -signy;
    if (signx != signy) return signx;
    // Now we know that a[0] and b[0] have the same sign.
    // Deal with cases where exponents differ by more than 1.
    long int expx, expy;
    mpf_get_d_2exp(&expx, a[0]);
    mpf_get_d_2exp(&expy, b[0]);
    if (expx > expy+1) return signx;
    if (expy > expx+1) return -signy;
    // Now we know that a[0] and b[0] have almost the same exponent.
    if (myIsEqualNZIgnoreSign(rawx, rawy)) return 0; // might throw InsufficientPrecision
    if (mpf_cmp(a[0], b[0]) > 0) return 1;
    return -1;
  }


  int RingTwinFloatImpl::mySign(ConstRawPtr rawx) const
  {
    return mpf_sgn(import(rawx)[0]);
  }


  //---------------------------------------------------------------------------

  // This function has to exist, but it can never be called,
  RingHom RingTwinFloatImpl::myCompose(const RingHom& phi, const RingHom& /*theta*/) const
  {
    CoCoA_ERROR(ERR::SERIOUS, "RingTwinFloatImpl::myCompose -- how did you get here?");
    return phi; // just to keep compiler quiet
  }


  ideal RingTwinFloatImpl::myIdealCtor(const std::vector<RingElem>& gens) const
  {
    // Is it wise to allow one to create an ideal in RingTwinFloat???
    return NewFieldIdeal(ring(this), gens);
  }


  void RingTwinFloatImpl::myCheckConsistency(RawPtr rawx) const
  {
//     std::cout<<"CHECK INPUT IS ";myOutput(std::cout,x);std::cout<<std::endl;

    MultipleFloat_t a(import(rawx));
    CoCoA_ASSERT(mpf_sgn(a[0]) != 0);
    for (size_t i = 1; i < myNumCompts; ++i)
    {
      CoCoA_ASSERT(mpf_sgn(a[i]) != 0);
      if (mpf_sgn(a[i]) != mpf_sgn(a[0])) throw RingTwinFloat::InsufficientPrecision();
      size_t accuracy = ApproximatelyEqual(a[0], a[i], mySNRBits+myNoiseBits/2);
//        std::cout<<"CHECK: For compt " << i << "  accuracy=" << accuracy<< std::endl;
      if (accuracy < myMantissaBits+myGuardBits)
      {
//       std::cout<<"FOUND INCONSISTENCY"<<std::endl;myOutput(std::cout,x);std::cout<<std::endl;
        throw RingTwinFloat::InsufficientPrecision();
      }
      if (accuracy >= mySNRBits+myNoiseBits/2)                 // if there is too little noise...
        perturb(a[i], mySNRBits, myNoiseBits, myRandomSource); // ...add some more.
    }
  }


  // Let delta be the value this function returns; then by definition
  // x is considered equal to all numbers whose first delta bits match
  // those of x; more precisely x is equal to all y for which the binary
  // binary exponent of |x-y| is less than the binary exponent of x minus delta.
  size_t RingTwinFloatImpl::myMatchBitsForEqual(ConstRawPtr /*rawx*/) const
  {
    // if fixed precision...
    return myMantissaBits+myGuardBits/2;
//    else...
//     // VARIABLE PRECISION CASE
//     const size_t ReprPrecision = myMantissaBits+myGuardBits+myNoiseBits;
//     return ApproximatelyEqual(import(rawx)[0], import(rawx)[1], ReprPrecision) - myGuardBits/2;

}

  // ASSUMES x and y are non-zero.
  // IGNORES signs of x and y; ignores the trivial case of exponents differing by more than.
  bool RingTwinFloatImpl::myIsEqualNZIgnoreSign(ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    CoCoA_ASSERT(!myIsZero(rawx) && !myIsZero(rawy));

    // ??? ANNA to keep the compiler quiet
//     ConstMultipleFloat_t a(import(rawx));
//     ConstMultipleFloat_t b(import(rawy));
//#ifdef CoCoA_DEBUG
//     {
//       // obviously unequal if the exponents differ by more than 1
//       long expx, expy;
//       mpf_get_d_2exp(&expx, import(rawx)[0]);
//       mpf_get_d_2exp(&expy, import(rawy)[0]);
//       CoCoA_ASSERT(expx <= expy+1 && expx+1 >= expy); // equiv.cond. abs(expx-expy) <= 1
//     }
//#endif

    const size_t MinimumMatch = min(myMatchBitsForEqual(rawx), myMatchBitsForEqual(rawy)); // two calls are redundant in fix prec case...

    const size_t MatchBits = ApproximatelyEqual(import(rawx)[0], import(rawy)[0], MinimumMatch);
    if (MatchBits >= MinimumMatch) return true;
    if (MatchBits < myMantissaBits) return false;
    throw RingTwinFloat::InsufficientPrecision();
  }


  RingTwinFloat::InsufficientPrecision::InsufficientPrecision():
      ErrorInfo(ERR::InsuffPrec, "RingTwinFloat arithmetic")
  {}


  //---------------------------------------------------------------------------

  bool IsPracticallyEqual(ConstRefRingElem x, ConstRefRingElem y)
  {
    if (owner(x) != owner(y))
      CoCoA_ERROR(ERR::MixedRings, "IsPracticallyEqual(RingElem,RingElem)");
    if (!IsRingTwinFloat(owner(x)))
      CoCoA_ERROR(ERR::NotRingTwinFloat, "IsPracticallyEqual(RingElem,RingElem)");

    try
    {
      return x == y;
    }
    catch (RingTwinFloat::InsufficientPrecision&)
    {
      return false;
    }
  }


  RingTwinFloat::RingTwinFloat(const RingTwinFloatBase* RingPtr):
      ring(RingPtr)
  {}


  RingTwinFloat NewRingTwinFloat(size_t MantissaBits)
  {
    return RingTwinFloat(new RingTwinFloatImpl(MantissaBits));
  }


  bool IsRingTwinFloat(const ring& RR)
  {
    return dynamic_cast<const RingTwinFloatImpl*>(RR.myRawPtr()) != 0;
  }


  RingTwinFloat AsRingTwinFloat(const ring& R)
  {
    if (!IsRingTwinFloat(R))
      CoCoA_ERROR(ERR::NotRingTwinFloat, "AsRingTwinFloat(ring)");
    return RingTwinFloat(static_cast<const RingTwinFloatImpl*>(R.myRawPtr()));
  }


  size_t RingTwinFloatImpl::myPrecisionBits() const
  {
    return myMantissaBits;
  }


  size_t PrecisionBits(const RingTwinFloat& RR)
  {
    return RR->myPrecisionBits();
  }


} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/RingTwinFloat.C,v 1.3 2007/10/30 17:14:07 abbott Exp $
// $Log: RingTwinFloat.C,v $
// Revision 1.3  2007/10/30 17:14:07  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.2  2007/09/24 14:24:41  abbott
// Added parentheses to make code more readable (and to shut up gcc-4.3).
//
// Revision 1.1  2007/06/21 21:29:47  abbott
// Changed name of RingFloat into RingTwinFloat.
//
// Revision 1.3  2007/05/22 22:45:14  abbott
// Changed fn name IsUnit to IsInvertible.
//
// Revision 1.2  2007/03/21 14:52:49  bigatti
// -- added AsRingFloat(R)
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.15  2007/03/08 18:22:29  cocoa
// Just whitespace cleaning.
//
// Revision 1.14  2007/03/07 14:07:39  bigatti
// -- minor: commented argument names for -Wextra
//
// Revision 1.13  2007/01/17 12:32:39  cocoa
// Changed a few more "raw" variable names so that the code compiles fine
// also when CoCoA_DEBUG is set.
//
// Revision 1.12  2007/01/15 14:31:35  bigatti
// -- added prefix "raw" to RawPtr arguments names
// -- changed rhs into rawx, n, or N
//
// Revision 1.11  2007/01/09 15:52:08  cocoa
// Changed QBGenerator to use std::vector instead of std::list for the result.
// Minor mod to configure script.
//
// Revision 1.10  2006/12/21 13:48:32  cocoa
// Made all increment/decrement calls prefix (except where the must be postfix).
//
// Revision 1.9  2006/12/06 17:36:58  cocoa
// -- commented out 2 lines to keep compiler quiet
//
// Revision 1.8  2006/11/27 14:25:53  cocoa
// -- reorganised #include files
//
// Revision 1.7  2006/11/27 13:06:22  cocoa
// Anna and Michael made me check without writing a proper message.
//
// Revision 1.6  2006/11/03 14:01:46  cocoa
// -- changed: reference counting in ring, PPMonoids and OrdvArith now
//    uses SmartPtrIRC
//
// Revision 1.5  2006/11/02 13:25:44  cocoa
// Simplification of header files: the OpenMath classes have been renamed.
// Many minor consequential changes.
//
// Revision 1.4  2006/10/16 23:18:59  cocoa
// Corrected use of std::swap and various special swap functions.
// Improved myApply memfn for homs of RingDistrMPolyInlPP.
//
// Revision 1.3  2006/10/06 14:04:14  cocoa
// Corrected position of #ifndef in header files.
// Separated CoCoA_ASSERT into assert.H from config.H;
// many minor consequential changes (have to #include assert.H).
// A little tidying of #include directives (esp. in Max's code).
//
// Revision 1.2  2006/07/17 19:40:53  cocoa
// Extensive changes following a slight variantion in the semantics.  Added fn IsRational.
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.7  2006/05/29 16:22:37  cocoa
// Third time lucky???
// Added myIsInteger member function to all rings (NYI for RingFloat).
//
// Revision 1.6  2006/05/12 16:10:58  cocoa
// Added OpenMathFwd.H, and tidied OpenMath.H.
// Many consequential but trivial changes.
//
// Revision 1.5  2006/03/15 18:09:31  cocoa
// Changed names of member functions which print out their object
// into myOutputSelf -- hope this will appease the Intel C++ compiler.
//
// Revision 1.4  2006/03/14 15:01:49  cocoa
// Improved the implementation of ring member fns for computing powers.
// Should keep Intel C++ compiler quieter too.
//
// Revision 1.3  2006/03/12 21:28:33  cocoa
// Major check in after many changes
//
// Revision 1.2  2006/02/20 22:41:19  cocoa
// All forms of the log function for power products now return SmallExponent_t
// (instead of int).  exponents now resizes the vector rather than requiring
// the user to pass in the correct size.
//
// Revision 1.1.1.1  2005/10/17 10:46:54  cocoa
// Imported files
//
// Revision 1.4  2005/10/11 16:37:30  cocoa
// Added new small prime finite field class (see RingFpDouble).
//
// Cleaned makefiles and configuration script.
//
// Tidied PPMonoid code (to eliminate compiler warnings).
//
// Fixed bug in RingFloat::myIsInteger.
//
// Revision 1.3  2005/09/22 18:04:17  cocoa
// It compiles; the tests run OK.  The examples compile.
// No documentation -- the mindless eurocrats have rendered
// me mindless too.
//
// Revision 1.2  2005/06/22 14:42:16  cocoa
// Renamed MemPool data member to myMemMgr
// (seems more sensible than myMemory).
//
// Revision 1.1.1.1  2005/05/03 15:47:31  cocoa
// Imported files
//
// Revision 1.4  2005/04/20 15:40:48  cocoa
// Major change: modified the standard way errors are to be signalled
// (now via a macro which records filename and line number).  Updated
// documentation in error.txt accordingly.
//
// Improved the documentation in matrix.txt (still more work to be done).
//
// Revision 1.3  2005/04/19 14:06:03  cocoa
// Added GPL and GFDL licence stuff.
//
// Revision 1.2  2005/02/11 14:15:20  cocoa
// New style ring elements and references to ring elements;
// I hope I have finally got it right!
//
// Revision 1.1.1.1  2005/01/27 15:12:13  cocoa
// Imported files
//
// Revision 1.13  2004/12/16 13:34:12  cocoa
// Fixed a memory leak, and noted some areas needing improvement.
//
// Revision 1.12  2004/11/18 18:33:41  cocoa
// Now every ring know its own "one" element (as well as "zero").
// Several consequential changes.
//
// Revision 1.11  2004/11/11 11:56:09  cocoa
// (1) Tidied makefiles, and introduced common.mki
// (2) Improved several tests, and cleaned them so that they
//     handle sanely any otherwise unhandled exceptions.
//
// Revision 1.10  2004/11/05 15:34:33  cocoa
// Consequential change following from the renaming of
// FieldIdealImpl and the introduction of the new pseudo-ctor.
//
// Revision 1.9  2004/11/04 18:47:43  cocoa
// (1) Ring member functions which previously expected mpz_t args
//     now expect ZZ args.  Numerous minor consequential changes.
// (2) Renamed function which gives access to the mpz_t value inside
//     a ZZ object: previously was raw(...), now is mpzref(...).
//     Plenty of calls had to be altered.
//
// Revision 1.8  2004/07/27 16:03:39  cocoa
// Added IsCommutative test and IamCommutative member function
// to all rings.  Tidied geobuckets a little.
//
// Revision 1.7  2004/05/27 16:14:02  cocoa
// Minor revision for new coding conventions.
//
// Revision 1.6  2004/05/24 15:52:13  cocoa
// Major update:
//   new error mechanism
//   many fixes
//   RingHoms almost work now
//   RingFloat much improved
//
// Revision 1.5  2004/04/08 15:33:34  cocoa
// Added function IsInteger, and the related RingBase::myIsInteger
// virtual function, plus all necessary implementations.
//
// Revision 1.4  2004/04/02 16:10:56  cocoa
// -- output: symbol "e" changed into "10^" to be read by cocoa-4
//
// Revision 1.3  2004/03/20 17:46:10  cocoa
// Check in prior to departure to RWCA
//
// Revision 1.2  2004/02/03 16:16:20  cocoa
// Removed pointless IamGCDDomain functions from several concrete rings.
// Added IamOrderedDomain functions where appropriate.
// Tidied ctors for the small finite fields.
//
// Revision 1.1  2004/01/28 15:54:09  cocoa
// Sundry additions.
//
