//   Copyright (c)  2006  John Abbott
//   Main authors: Laura Torrente (assisted by John Abbott)

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/ApproxPts.H"
#include "CoCoA/assert.H"
#include "CoCoA/CanonicalHom.H"
#include "CoCoA/convert.H"
#include "CoCoA/DenseMatrix.H"
#include "CoCoA/FractionField.H"
#include "CoCoA/io.H"
#include "CoCoA/matrix.H"
#include "CoCoA/MatrixArith.H"
#include "CoCoA/MatrixViews.H"
#include "CoCoA/PolyRing.H"
#include "CoCoA/PPMonoid.H"
#include "CoCoA/QBGenerator.H"
#include "CoCoA/ring.H"
#include "CoCoA/RingHom.H"
#include "CoCoA/RingQ.H"
#include "CoCoA/RingTwinFloat.H"
#include "CoCoA/SparsePolyRing.H"
#include "CoCoA/symbol.H"

#include <iostream>
using std::ostream;
using std::endl;

//#include <vector>
using std::vector;
#include <list>
using std::list;
#include <algorithm>
using std::copy;
using std::find_if;
#include <cmath>
using std::log; // only in UnitVolBallRadius
using std::exp; // only in UnitVolBallRadius

namespace CoCoA
{

  namespace ApproxPts
  {

    namespace // anonymous namespace for file local auxiliary functions and definitions.
    {

      // This is not really specific to approximate points, but where else should it go?
      // Compute radius of D-ball having volume 1.
      // (I got the formula from Wikipedia under "hypersphere")
      double UnitVolBallRadius(size_t D)
      {
        CoCoA_ASSERT(D < 10000); // disallow very high dimensions
        const double pi = 4*atan(1.0);
        double logfactor = (D/2)*(-std::log(2*pi)); // D/2 exploits integer division!  NB -(D/2)*log(...) does NOT WORK.
        if (D & 1)
          logfactor -= std::log(2.0);
        for (size_t k=D; k > 1; k -= 2)
          logfactor += std::log(double(k));
        return std::exp(logfactor/D);
      }


      // Compute square of 2-norm between P and Q.
      double L2NormSq(const ApproxPt& P, const ApproxPt& Q)
      {
        CoCoA_ASSERT(P.size() == Q.size());
        double SumSq = 0;
        const size_t dim = P.size();
        for (size_t i=0; i < dim; ++i)
          SumSq += square(P[i]-Q[i]);
        return SumSq;
      }


      // Data structure for representing PreProcessed Points.
      struct PPP
      {
        ApproxPt myCoords;      // the coords of the preprocessed point
        size_t myNumOrigPts;    // length of list myOrigPts (used only in PreprocessSubdivAlgm)
        list<size_t> myOrigPts; // indices into a master table
        list<double> myDistanceSq;  // used only in PreprocessSubdivAlgm
        double mySumDistanceSq;     // used only in PreprocessSubdivAlgm
      };


      // Returns the average of a set of ApproxPts; the set is indicated by the first arg
      // which is a list of indices into the second arg.
      ApproxPt ComputeCofG(const list<size_t>& IndexList, const vector<ApproxPt>& OrigPts)
      {
        const size_t dim = OrigPts[0].size();
        ApproxPt CofG(dim);
        for (list<size_t>::const_iterator it=IndexList.begin(); it != IndexList.end(); ++it)
        {
          const ApproxPt& P = OrigPts[*it];
          for (size_t i=0; i < dim; ++i)
          {
            CofG[i] += P[i];
          }
        }
        const size_t n = IndexList.size();
        for (size_t i=0; i < dim; ++i)
          CofG[i] /= n;
        return CofG;
      }


      // Returns true if the values are sane; false otherwise.
      bool CheckSemiWidths(const vector<double>& semiwidth)
      {
        const size_t NumDims = semiwidth.size();
        for (size_t i=0; i < NumDims; ++i)
          if (semiwidth[i] <= 0) return false;
        return true;
      }


      // Rescale coords of pt by factors in array ScaleFactor.
      void ScalePt(ApproxPt& pt, const vector<double>& ScaleFactor)
      {
        const size_t dim = ScaleFactor.size();
        for (size_t i=0; i < dim; ++i)
        {
          pt[i] *= ScaleFactor[i];
        }
      }

      // Rescale in place the OrigPts.
      void NormalizeOrigPts(vector<ApproxPt>& OrigPts, const vector<double>& semiwidth)
      {
        const size_t NumPts = OrigPts.size();
        const size_t dim = semiwidth.size();

        vector<double> InverseSemiwidth(dim);
        for (size_t j=0; j < dim; ++j)
          InverseSemiwidth[j] = 1/semiwidth[j];

        for (size_t i=0; i < NumPts; ++i)
          ScalePt(OrigPts[i], InverseSemiwidth);
      }


      /////////////////////////////////////////////////////////////////////////////
      // Auxiliary definitions for the grid algorithm.

      ApproxPt GridNearPoint(const ApproxPt& P)
      {
        const size_t dim = P.size();
        ApproxPt GNP(dim);
        for (size_t i=0; i < dim; ++i)
          GNP[i] = 2*round(P[i]/2);
        return GNP;
      }


      // Only used in GridAlgm as arg to std::find.
      // WARNING data member is a reference!!
      class CoordsEq
      {
      public:
        CoordsEq(const ApproxPt& P): myApproxPt(P) {};
        bool operator()(const PPP& PPPt) { return myApproxPt == PPPt.myCoords; };
      private: // data members
        const ApproxPt& myApproxPt;
      };


      /////////////////////////////////////////////////////////////////////////////
      // Functions for aggregative algorithm

      // ave = (w1*P1 + w2*P2)/(w1+w2);
      void WeightedAve(ApproxPt& ave, double w1, const ApproxPt& P1, double w2, const ApproxPt& P2)
      {
        const size_t dim = ave.size();
        CoCoA_ASSERT(P1.size() == dim && P2.size() == dim);
        CoCoA_ASSERT(w1+w2 != 0);
        for (size_t i=0; i < dim; ++i)
        {
          ave[i] = (w1*P1[i]+w2*P2[i])/(w1+w2);
        }
      }


      struct NearPair
      {
      public:
        NearPair(size_t i, size_t j, double sepsq): myLowerIndex(i), myHigherIndex(j), mySepSq(sepsq) {};
      public: // data members
        const size_t myLowerIndex, myHigherIndex;
        double mySepSq;
      };


      // Used only as arg to std::sort algorithm
      inline bool LessNearPair(const NearPair& P1, const NearPair& P2)
      {
        return P1.mySepSq < P2.mySepSq;
      }


      // Used in call to algorithm std::remove_if.
      class NearPairContains
      {
      public:
        NearPairContains(size_t j): myIndex(j) {}
        bool operator()(const NearPair& P) { return P.myLowerIndex == myIndex || P.myHigherIndex == myIndex; }
      private: // data member
        const size_t myIndex;
      };




      /////////////////////////////////////////////////////////////////////////////
      // Below is code for the "subdiv" algorithm.


      // Fill in data members myDistanceSq and mySumDistanceSq of X
      void ComputeDistSq(PPP& X, const vector<ApproxPt>& OrigPts)
      {
        X.myDistanceSq.clear();
        double SumDistanceSq = 0;
        const list<size_t>::const_iterator finished = X.myOrigPts.end();
        for (list<size_t>::const_iterator it = X.myOrigPts.begin(); it != finished; ++it)
        {
          double DistanceSq = L2NormSq(X.myCoords, OrigPts[*it]);
          X.myDistanceSq.push_back(DistanceSq);
          SumDistanceSq += DistanceSq;
        }
        X.mySumDistanceSq = SumDistanceSq;
      }


      // Make Y a copy of X but without the i-th OrigPt.
      void EraseAndUpdate(PPP& Y, const PPP& X, size_t i, const vector<ApproxPt>& V)
      {
        Y.myOrigPts = X.myOrigPts;
        Y.myOrigPts.remove(i);
        Y.myNumOrigPts = X.myNumOrigPts-1;
        Y.myCoords = ComputeCofG(Y.myOrigPts, V);
        ComputeDistSq(Y, V);
      }


      // Make Y a copy of X with the i-th OrigPt added in.
      void AddAndUpdate(PPP& Y, const PPP& X, size_t i, const vector<ApproxPt>& V)
      {
        Y.myOrigPts = X.myOrigPts;
        Y.myOrigPts.push_back(i);
        Y.myNumOrigPts = X.myNumOrigPts+1;
        Y.myCoords = ComputeCofG(Y.myOrigPts, V);
        ComputeDistSq(Y, V);
      }


      // Redistribute the assignments of OrigPts to PPPts to minimize the sum of squares of distances
      void redistribute(vector<PPP>& PPPts, const vector<ApproxPt>& OrigPts)
      {
        const size_t NumPPP = PPPts.size();
        bool MadeAnImprovement = true;
        while (MadeAnImprovement)
        {
          MadeAnImprovement = false;
          double MaxGain = 0.000001; // morally zero but must allow for rounding error
          size_t index_from = 0; // index of PPPt we move from (initialize to keep compiler quiet)
          size_t index_to = 0;   // index of PPPt we move to (initialize to keep compiler quiet)
          size_t IndexOrigPt = 0; // index of orig point we want to move (initialize to keep compiler quiet)

          // For each PPPt i do ...
          for (size_t i=0; i < NumPPP; ++i)
          {
            const double Ni = double(PPPts[i].myNumOrigPts);
            if (Ni == 1) continue;
            const double coeffi = Ni/(Ni-1);
            // For each index (of OrigPt associated to PPPt i) it do...
            list<double>::iterator it_dist = PPPts[i].myDistanceSq.begin();
            for (list<size_t>::iterator it = PPPts[i].myOrigPts.begin(); it != PPPts[i].myOrigPts.end(); ++it, ++it_dist)
            {
              const double diffi = (*it_dist)*coeffi;

              // For each PPPt j (difft from PPPt i) do...
              for (size_t j=0; j < NumPPP; ++j)
              {
                if (j == i) continue;
                const double Nj = double(PPPts[j].myNumOrigPts);
                const double coeffj = Nj/(Nj+1);
                const double diffj = -coeffj * L2NormSq(PPPts[j].myCoords, OrigPts[*it]);
                const double gain = diffi + diffj;
                if (gain > MaxGain)
                {
                  MadeAnImprovement = true;
                  MaxGain = gain;
                  index_from = i;
                  index_to = j;
                  IndexOrigPt = *it;
                }
              }
            }
          }

          if (MadeAnImprovement)
          {
            // Move orig point to new PPPt...
            AddAndUpdate(PPPts[index_to], PPPts[index_to], IndexOrigPt, OrigPts);
            EraseAndUpdate(PPPts[index_from], PPPts[index_from], IndexOrigPt, OrigPts);
          }
        }
      }


      // Initialize the list of preprocessed points for subdivision algm.
      void PreprocessSubdivInit(vector<PPP>& PPPts, const vector<ApproxPt>& OrigPts)
      {
        CoCoA_ASSERT(PPPts.empty());
        PPPts.resize(1);
        const size_t NumPts = OrigPts.size();
        for (size_t i=0; i < NumPts; ++i)
          PPPts[0].myOrigPts.push_back(i);
        PPPts[0].myNumOrigPts = NumPts;
        PPPts[0].myCoords = ComputeCofG(PPPts[0].myOrigPts, OrigPts);
        ComputeDistSq(PPPts[0], OrigPts);
      }


      // Determine if some OrigPt is too far from its corresponding PPPt.
      // If result is true, then WorstOrigIdx and WorstPPPIdx contain the indices
      // of the most distant pair of orig and preprocessed points.
      bool ExistsBadPoint(size_t& WorstOrigIdx, size_t& WorstPPPIdx, list<size_t>::iterator& it_list_worst, /*const*/ vector<PPP>& V)
      {
        it_list_worst = V[0].myOrigPts.begin();
        list<double>::iterator it_dist_worst = V[0].myDistanceSq.begin();
        const size_t NumPPP = V.size();
        WorstPPPIdx = 0;
        for (size_t i=0; i < NumPPP; ++i)
        {
          list<size_t>::iterator it_list = V[i].myOrigPts.begin();
          for (list<double>::iterator it_dist = V[i].myDistanceSq.begin(); it_dist != V[i].myDistanceSq.end(); ++it_dist)
          {
            if (*it_dist > *it_dist_worst)
            {
              it_dist_worst = it_dist;
              it_list_worst = it_list;
              WorstPPPIdx = i;
            }
            ++it_list;
          }
        }
        const size_t dim = V[0].myCoords.size();
        const double CriticalRadius = 2*UnitVolBallRadius(dim); // inefficient if dim is large and num of pts is small
        if (*it_dist_worst < CriticalRadius) return false;
        WorstOrigIdx = *it_list_worst;
        return true;
      }


      // Adjoin a new preprocessed point onto the end of V.
      // The new point comprises just one OrigPt, the one with index OrigPtIdx.
      void NewPPP(vector<PPP>& PPPts, const ApproxPt& P, size_t OrigPtIdx)
      {
        const size_t NumPPPts = PPPts.size();
        PPPts.resize(NumPPPts+1);
        PPP& NewPt(PPPts[NumPPPts]);
        NewPt.myCoords = P;
        NewPt.myOrigPts.push_back(OrigPtIdx);
        NewPt.myNumOrigPts = 1;
        NewPt.myDistanceSq.push_back(0);
        NewPt.mySumDistanceSq = 0;
      }


    } // end of anonymous namespace


    /////////////////////////////////////////////////////////////////////////////
    // Below are the three main algorithms.
    // First the "grid" algorithm.

    void PreprocessGridAlgm(vector<ApproxPt>& NewPts, vector<ApproxPt> OrigPts, const vector<double>& semiwidth)
    {
      const size_t NumPts = OrigPts.size();

      // Some sanity checks on the inputs.
      CoCoA_ASSERT(NumPts > 0);
      CoCoA_ASSERT(OrigPts[0].size() > 0);
      // Should check that all points lie in the same space.
      CoCoA_ASSERT(semiwidth.size() == OrigPts[0].size());
      CoCoA_ASSERT(CheckSemiWidths(semiwidth)); // give real error???

      NormalizeOrigPts(OrigPts, semiwidth);


      vector<PPP> PPPts; PPPts.reserve(NumPts);
      // For each OrigPts[i] do...
      for (size_t i = 0; i < NumPts; ++i)
      {
        const ApproxPt P = GridNearPoint(OrigPts[i]);
        vector<PPP>::iterator pos = find_if(PPPts.begin(), PPPts.end(), CoordsEq(P));
        if (pos != PPPts.end())
          pos->myOrigPts.push_back(i);         // adjoin to existing class
        else
          NewPPP(PPPts, P, i);  // add a new class
      }

      const size_t NumPPPts = PPPts.size();
      NewPts.resize(NumPPPts);
      for (size_t k = 0; k < NumPPPts; ++k)
      {
        NewPts[k] = ComputeCofG(PPPts[k].myOrigPts, OrigPts);
        ScalePt(NewPts[k], semiwidth);
      }

    }


    /////////////////////////////////////////////////////////////////////////////
    // The "aggregative" algorithm.

    void PreprocessAggrAlgm(vector<ApproxPt>& NewPts, vector<ApproxPt> OrigPts, const vector<double>& semiwidth)
    {
      const size_t NumPoints = OrigPts.size();
      const size_t dim = OrigPts[0].size();

      // Some sanity checks on the inputs.
      CoCoA_ASSERT(NumPoints > 0);
      CoCoA_ASSERT(dim > 0);
      CoCoA_ASSERT(semiwidth.size() == dim);
      CoCoA_ASSERT(CheckSemiWidths(semiwidth)); // give real error???

      // Normalization effectively makes the semiwidth in each direction equal to 1.
      NormalizeOrigPts(OrigPts, semiwidth);
      const double CriticalRadiusSq = square(2*UnitVolBallRadius(dim));

      vector<PPP> PreprocessedPoints(NumPoints);
      for (size_t i=0; i < NumPoints; ++i)
      {
        PreprocessedPoints[i].myCoords = OrigPts[i];
        PreprocessedPoints[i].myOrigPts.push_back(i);
      }

      // nbrs contains all pairs which could conceivably collapse together
      // (i.e. only pairs separated by at most the CriticalRadius)
      list<NearPair> nbrs;
      for (size_t i=0; i < NumPoints; ++i)
      {
        for (size_t j=i+1; j < NumPoints; ++j)
	{
	  const double d2 = L2NormSq(PreprocessedPoints[i].myCoords,
                                     PreprocessedPoints[j].myCoords);
	  if (d2 < 4*CriticalRadiusSq)
	    nbrs.push_back(NearPair(i,j,d2));

	}
      }

      nbrs.sort(LessNearPair);

      ApproxPt CofG(dim);
      while (!nbrs.empty() && nbrs.begin()->mySepSq < 4*CriticalRadiusSq)
      {
        const size_t i = nbrs.front().myLowerIndex;
        const size_t j = nbrs.front().myHigherIndex;
        //cout << "pair is " << i << "  " << j << endl;
        //cout << "dist=" << nbrs.front().mySepSq << endl;
        nbrs.erase(nbrs.begin()); // delete first elem of nbrs

        const PPP& Pi = PreprocessedPoints[i];
        const PPP& Pj = PreprocessedPoints[j];
        // Compute CofG of the union of Pi an Pj...
        WeightedAve(CofG, Pi.myOrigPts.size(), Pi.myCoords, Pj.myOrigPts.size(), Pj.myCoords);

        // Check that all original points in the union are close to CofG.
        bool OK = true;
        for (list<size_t>::const_iterator it=Pi.myOrigPts.begin(); OK && it != Pi.myOrigPts.end(); ++it)
	{
	  OK &= (L2NormSq(CofG, OrigPts[*it]) < CriticalRadiusSq);
	}
        for (list<size_t>::const_iterator it=Pj.myOrigPts.begin(); OK && it != Pj.myOrigPts.end(); ++it)
	{
	  OK &= (L2NormSq(CofG, OrigPts[*it]) < CriticalRadiusSq);
	}
        if (!OK) continue; // We cannot unite these two points, so go on to next pair of points.

        // The union is fine, so accept it.  We put all data into i-th PPP
        // and empty the data of the j-th PPP.
        swap(CofG, PreprocessedPoints[i].myCoords);
        {
          list<size_t>& PPPOrigPts = PreprocessedPoints[i].myOrigPts;
          PPPOrigPts.splice(PPPOrigPts.begin(), PreprocessedPoints[j].myOrigPts);
        }
        nbrs.remove_if(NearPairContains(j)); // since Pj has been eliminated.
        // Update all entries refering to i-th point.
        for (list<NearPair>::iterator nbr = nbrs.begin(); nbr != nbrs.end(); ++nbr)
	{
	  if (!NearPairContains(i)(*nbr)) continue;
	  size_t other = nbr->myLowerIndex;
	  if (other == i) other = nbr->myHigherIndex;
	  nbr->mySepSq = L2NormSq(CofG, PreprocessedPoints[other].myCoords);
	}
        nbrs.sort(LessNearPair); // put the list back into order -- SLUG SLUG SLUG!!!
        //cout << "len nbrs = " << nbrs.size() << endl;
      }

      // There are no futher possible mergings, so copy answer into NewPts.
      NewPts.clear();
      for (size_t i=0; i < NumPoints; ++i)
      {
        if (!PreprocessedPoints[i].myOrigPts.empty())
        {
          NewPts.push_back(PreprocessedPoints[i].myCoords);
          ScalePt(NewPts.back(), semiwidth);
        }
      }
    }


    /////////////////////////////////////////////////////////////////////////////
    // The "subdivision" algorithm.

    void PreprocessSubdivAlgm(vector<ApproxPt>& NewPts, vector<ApproxPt> OrigPts, const vector<double>& semiwidth)
    {
      const size_t NumPoints = OrigPts.size();

      // Some sanity checks on the inputs.
      CoCoA_ASSERT(NumPoints > 0);
      CoCoA_ASSERT(OrigPts[0].size() > 0);
      CoCoA_ASSERT(semiwidth.size() == OrigPts[0].size());
      CoCoA_ASSERT(CheckSemiWidths(semiwidth)); // give real error???

      // Normalization effectively makes the semiwidth in each direction equal to 1.
      NormalizeOrigPts(OrigPts, semiwidth);

      // Initially all orig points are associated to a single preprocessed point.
      vector<PPP> PPPoints; PPPoints.reserve(NumPoints);
      PreprocessSubdivInit(PPPoints, OrigPts);

      // Main loop.
      size_t WorstPPPIdx, WorstOrigIdx; // these are set in call to ExistsBadPoint
      list<size_t>::iterator it_worst;
      while (ExistsBadPoint(WorstOrigIdx, WorstPPPIdx, it_worst, PPPoints))
      {
        NewPPP(PPPoints, OrigPts[WorstOrigIdx], WorstOrigIdx);
        EraseAndUpdate(PPPoints[WorstPPPIdx], PPPoints[WorstPPPIdx], *it_worst, OrigPts);
        redistribute(PPPoints, OrigPts);
      }

      // Copy and denormalize result.
      NewPts.clear();
      NewPts.reserve(PPPoints.size());
      for (size_t i=0; i < PPPoints.size(); ++i)
      {
        NewPts.push_back(PPPoints[i].myCoords);
        ScalePt(NewPts[i], semiwidth);
      }

    }


    /////////////////////////////////////////////////////////////////////////////

typedef vector<RingElem> point;


/// Convert the input data to the RingTwinFloat with current BitPrec 
// !!! PROTECT AGAINST ALIASING!!!
void ConversionToRR(const vector<RingElem>& OrigSemiwidth, 
                    const vector<point>& OrigPts, 
		    ConstRefRingElem OrigGamma,
                    vector<RingElem>& semiwidth, 
                    vector<point>& Pts, 
		    RingElem& Gamma,
                    const ring& RR)
{
  ring Q = owner(OrigSemiwidth.front());
  RingHom QToRR = CanonicalHom(Q,RR);
  semiwidth.clear(); Pts.clear();
  for (size_t i=0; i < OrigSemiwidth.size(); ++i)
    semiwidth.push_back(QToRR(OrigSemiwidth[i]));
  for (size_t i=0; i < OrigPts.size(); ++i)
    {
      point NewPt;
      for (size_t j=0; j < OrigSemiwidth.size(); ++j)
	NewPt.push_back(QToRR(OrigPts[i][j]));
      Pts.push_back(NewPt);
    }
  Gamma = QToRR(OrigGamma);
}

void ConversionToRR(const vector<double>& OrigSemiwidth, 
                    const vector<ApproxPt>& OrigPts, 
		    double OrigGamma,
                    vector<RingElem>& semiwidth, 
                    vector<point>& pts, 
		    RingElem& gamma,
                    const ring& RR)
{
  ZZ N,D; // used for converting doubles into exact rationals
  semiwidth.clear();
  for (size_t i=0; i < OrigSemiwidth.size(); ++i)
    {
      convert(N,D,OrigSemiwidth[i]);
      semiwidth.push_back(RingElem(RR,N)/D);
    }
  pts.clear();
  for (size_t j=0; j < OrigPts.size(); ++j)
    {
      vector<RingElem> tmp;
      for (size_t i=0; i < OrigPts[j].size(); ++i)
	{
	  convert(N,D, OrigPts[j][i]);
	  tmp.push_back(RingElem(RR,N)/D);
	}
      pts.push_back(tmp);
    }
  convert(N,D, OrigGamma);
  gamma = RingElem(RR,N)/D;
}


matrix MatrixEmbedding(ConstMatrixView M, const PolyRing& P)
{
  //ring R = BaseRing(M);
  //ERROR : If (CoeffRing(P) != R)   then error!!!
  //ERROR : If they have different size   then error!!
  matrix NewM(NewDenseMat(P, NumRows(M), NumCols(M)));
  RingHom phi = CoeffEmbeddingHom(P);
  for (size_t i=0; i < NumRows(M); ++i)
    for (size_t j=0; j < NumCols(M); ++j)
      SetEntry(NewM, i, j, phi(M(i,j)));
  return NewM;
}


vector<RingElem> VectorEmbedding(const vector<RingElem>& V, 
                                 const PolyRing& P)
{
  //ring R = owner(V.front());
  //ERROR : If (CoeffRing(P) != R)   then error!!!
  RingHom phi = CoeffEmbeddingHom(P);
  vector<RingElem> NewV;
  for (size_t i=0; i < V.size(); ++i)
    NewV.push_back(phi(V[i]));
  return NewV;
}


// square 2-norm of a vector
RingElem squareNorm(const vector<RingElem>& v)
{
  RingElem tmpSqNorm(zero(owner(v.front())));
  for (size_t i=0; i < v.size(); ++i)
    tmpSqNorm += power(v[i], 2);
  return tmpSqNorm;
}


// Evaluation of a poly f on a set Pts of points
void Evaluation(vector<RingElem>& b, 
                ConstRefRingElem f, 
                const list<RingHom>& EvalAtPts)
{
  b.clear();
  for (list<RingHom>::const_iterator it = EvalAtPts.begin(); it != EvalAtPts.end(); ++it)
    b.push_back((*it)(f));
}


void FirstOrderEval(vector<RingElem>& V,
                    ConstRefRingElem t, 
                    const list<RingHom>& EvalAtPts,
                    const vector<RingElem>& semiwidth,
                    const PolyRing& PErr)
{
  RingHom phi = CoeffEmbeddingHom(PErr);
  const size_t dim = semiwidth.size();
  const size_t NumPts = EvalAtPts.size();
  vector<RingElem> tmp;
  const vector<RingElem>& e = indets(PErr); 
  
  V.assign(NumPts, zero(PErr));
  for (size_t i=0; i < dim; ++i)
    {
      if (!IsZero(semiwidth[i])) 
	{
	  Evaluation(tmp, deriv(t,i), EvalAtPts);
	  for (size_t j=0; j < NumPts; ++j)
	    V[j] += phi(tmp[j]) * e[j*dim + i];
	}
    }
}


// Solve the LS problem A*x = b, A orthogonal matrix
// implemented without deleting the denominators 
void LeastSquares(vector<RingElem>& x,
                  vector<RingElem>& Rho, 
                  ConstMatrixView A,
                  ConstMatrixView DiagA, 
                  const vector<RingElem>& b) 
{
  ring R = owner(b.front());
  const size_t n = b.size();
  vector<RingElem> RhoRho(n, zero(R));
  
  x = vector<RingElem>(NumRows(A),zero(R));
  A->myMulByCol(x,b);
  for (size_t i=0; i < n; ++i)
    if (DiagA(i,i) != 0)
      x[i] /= DiagA(i,i); 
    else
      x[i] = zero(R);
  A->myMulByRow(RhoRho,x);
  Rho = b;
  for (size_t i = 0; i < n; ++i)
    Rho[i] -= RhoRho[i];
} 


// First order least squares algorithm
void FirstOrderLS(vector<RingElem>& Rho1,
                  vector<RingElem>& x1,
                  ConstMatrixView A0,
                  ConstMatrixView A1, 
                  ConstMatrixView DiagA0,
                  ConstMatrixView A0A1, 
                  const vector<RingElem>& b0,
                  const vector<RingElem>& b1, 
                  const vector<RingElem>& x0)
{
  SparsePolyRing P = AsSparsePolyRing(owner(b1.front()));
  RingHom phi = CoeffEmbeddingHom(P);
  matrix A0InP = MatrixEmbedding(A0, P);
  const vector<RingElem> x0InP = VectorEmbedding(x0, P);

  // x1 = A0*b1;
  x1.assign(NumRows(A0), zero(P));
  A0InP->myMulByCol(x1,b1);
  // B = A1*b0;
  vector<RingElem> B(NumRows(A1), zero(P));
  A1->myMulByCol(B, VectorEmbedding(b0, P));
  // C = A0A1*x0;
  vector<RingElem> C(NumRows(A0A1), zero(P));
  A0A1->myMulByCol(C,x0InP);
  for (size_t i=0; i < b1.size(); ++i)
    {
      if (IsZero(DiagA0(i,i)))
	x1[i] = zero(P);
      else 
	{
	  x1[i] += (B[i] - C[i]);
	  x1[i] *= phi(1/DiagA0(i,i));
        }
    }
  Rho1 = b1;
  // D = A0*x1;
  vector<RingElem> D(NumRows(A0A1), zero(P));
  A0InP->myMulByRow(D,x1);
  // E =  A1*x0;
  vector<RingElem> E(NumRows(A1), zero(P));
  A1->myMulByRow(E,x0InP);
  for (size_t i=0; i < b1.size(); ++i)
    Rho1[i] -= (D[i] + E[i]);
}


// Compute the matrix C that represents the vector V
void ErrorMatrix(MatrixView& C, const vector<RingElem>& V)
{
  SparsePolyRing P = AsSparsePolyRing(owner(V.front()));
  RingHom PToR = EvalHom(P, vector<RingElem>(NumIndets(P), zero(CoeffRing(P))));
  for (size_t i=0; i < V.size(); ++i)
    for (size_t j=0; j < NumIndets(P); ++j)
      SetEntry(C,i,j,PToR(deriv(V[i],j)));
}


// //Compute an estimate of the squared minimal singular value of a matrix 
// void LowerBoundMinSingValue(RingElem& LowerBound,
//                             ConstMatrixView A)
// { 
//   list<RingElem> listOfNorms;
//   RingElem norm(BaseRing(A));
//   inftyMatrixNorm(norm, A);
//   listOfNorms.push_back(1/(NumRows(A)*norm*norm));
//   GlobalOutput() << "LB1 = " << 1/(NumRows(A)*norm*norm) << endl;
//   inftyMatrixNorm(norm, transpose(A));
//   listOfNorms.push_back(1/(NumCols(A)*norm*norm));
//   GlobalOutput() << "LB2 = " << 1/(NumCols(A)*norm*norm) << endl;
//   FrobeniusNorm(norm, A);
//   listOfNorms.push_back(1/norm);
//   GlobalOutput() << "LB3 = " << 1/norm << endl;
//   listOfNorms.sort();
//   list<RingElem>::iterator it = listOfNorms.end();
//   LowerBound = *(--it);
//   GlobalOutput() << "Lower Bound is = " << LowerBound << endl;
// }



// Compute the vector of rows of the matrix A such that the minimal 
// singular value of the corresp submatrix is greater than ||Eps||_2
void WellIndepRows(vector<size_t>& rows, 
                   ConstRefRingElem Eps_square, 
                   ConstMatrixView A)
{
  const ring R = BaseRing(A);
  vector<size_t> cols;
  for (size_t j=0; j < NumCols(A); ++j)
    cols.push_back(j);
  rows.empty();
  for (size_t i=0; i < NumRows(A); ++i)
  {
    rows.push_back(i);
    ConstMatrixView B = submat(A, rows, cols);
    // Check lower bound for smallest singular value of B
    if (rank(B) < NumRows(B) || 1/FrobeniusNorm2(PseudoInverse(B)) < Eps_square)
    {
      rows.pop_back();
    }
  }
}



//Compute minimal 2-norm solution of a Full Rank Underdetermined System
void SolveFRUDS(vector<RingElem>& x,
                ConstMatrixView A, 
                const vector<RingElem>& b)
{
  ConstMatrixView PseudoInvA = PseudoInverse(A);
  // x = PseudoInvA*b;
  x.assign(NumRows(PseudoInvA), zero(BaseRing(A)));
  PseudoInvA->myMulByCol(x, b);
}



//Add to QB and update
void AddToQBAndUpdate(const PPMonoidElem& t, 
		      QBGenerator& QBG, 
		      const vector<RingElem>& Alpha0, 
		      vector<RingElem>& S)
{
  SparsePolyRing P = AsSparsePolyRing(owner(S.front()));
  RingHom RToP = CoeffEmbeddingHom(P);
  const RingElem tAsPoly = monomial(P, one(CoeffRing(P)), t);

  GlobalLogput() << "The term " << t << " is added to QB" << endl;
  const size_t lenQB = (QBG.myQB()).size();
  S.push_back(tAsPoly);
  for (size_t i=0; i < lenQB; ++i)
    S[lenQB] -= RToP(Alpha0[i])*S[i];
  QBG.myCornerPPIntoQB(t);
}


//Update evaluation matrices
void UpdateMatrices(const QBGenerator& QBG,
		    MatrixView& A0,
		    MatrixView& A1,
		    MatrixView& DiagA0,
		    MatrixView& A0A1,
		    const vector<RingElem>& Rho0,
		    const vector<RingElem>& Rho1)
{
  PolyRing PErr = AsPolyRing(BaseRing(A1));
  RingHom RRToPErr = CoeffEmbeddingHom(PErr);
  const size_t k = (QBG.myQB()).size();
  SetEntry(DiagA0, k, k, squareNorm(Rho0));
  for (size_t j=0; j < Rho0.size(); ++j)
    {
      SetEntry(A0, k, j, Rho0[j]);
      SetEntry(A1, k, j, Rho1[j]);
      SetEntry(A0A1, k, k, A0A1(k, k) + 2*Rho1[j]*RRToPErr(Rho0[j]));
    }
}


//Add to Corner set and update
void AddToCornerAndUpdate(const PPMonoidElem& t, 
			  QBGenerator& QBG, 
			  const vector<RingElem>& Alpha0, 
			  const vector<RingElem>& S, 
			  vector<RingElem>& J)
{
  SparsePolyRing P = AsSparsePolyRing(owner(S.front()));
  RingHom RToP = CoeffEmbeddingHom(P);
  const RingElem tAsPoly = monomial(P, one(CoeffRing(P)), t);

  GlobalLogput() << "The term " << t << " is added to Corner" << endl;
  RingElem f(tAsPoly);
  for (size_t i=0; i < S.size(); ++i)
    f -= RToP(Alpha0[i])*S[i];
  J.push_back(f);
  QBG.myCornerPPIntoAvoidSet(t);
}



// Compute the (first order) border of an order ideal
// controllare il nome, mettere controllo su O = vuoto
void ComputeBorder(list<PPMonoidElem>& Border, 
                   const QBGenerator& QBG)
{
  const vector<PPMonoidElem> QB = QBG.myQB();
  if (QB.empty()) 
    CoCoA_ERROR("Empty QB", "ComputeBorder");
  const vector<PPMonoidElem>& X = indets(owner(QB.front()));
  Border.clear();
  for (size_t i=0; i < QB.size(); ++i)
    for (size_t j=0; j < X.size(); ++j)
      Border.push_back(QB[i]*X[j]);
  Border.sort();
  Border.unique();
  for (size_t i=0; i < QB.size(); ++i)
    Border.remove(QB[i]);
}


//Compute the poly f = t-sum x_1 S_i
void ComputeBorderPoly(RingElem& f,
		       const PPMonoidElem& t,
		       ConstMatrixView A,
		       ConstMatrixView DiagA,
		       const list<RingHom>& EvalAtPts,
		       const vector<RingElem>& S)
{
  SparsePolyRing P = AsSparsePolyRing(owner(S.front()));
  RingHom phi = CoeffEmbeddingHom(P);
  f = monomial(P, 1, t);
  vector<RingElem> b, x, Rho;
  Evaluation(b, f, EvalAtPts);
  LeastSquares(x, Rho, A, DiagA, b);
  for (size_t i=0; i < S.size(); ++i)
    f -= phi(x[i])*S[i];
}


//Testing if a set of polys vanishing on Pts
bool IsBorderBasis(const vector<RingElem>& BBasis,
		   const list<RingHom>& EvalAtPts)
{
  for (size_t i=0; i < BBasis.size(); ++i)
    {
      vector<RingElem> b;
      Evaluation(b, BBasis[i], EvalAtPts);
      if (!IsZero(b)) 
	return false;
    }
  return true;	      
}


void SOI(std::vector<PPMonoidElem>& QB, std::vector<RingElem>& BBasis, std::vector<RingElem>& AlmostVanishing, const std::vector<ApproxPt>& OrigPts, const std::vector<double>& OrigSemiwidth, double OrigGamma)
{
  const ring Q = RingQ();
  const size_t dim = OrigSemiwidth.size();
  const size_t NumPts = OrigPts.size();

  PPOrdering PPO = NewStdDegRevLexOrdering(dim);
  PPMonoid PPM = NewPPMonoid(SymbolRange("x", 0, dim-1), PPO);
    
  for (size_t BitPrec = 512; ; BitPrec *= 2)
  {
    try
    {
      ring RR = NewRingTwinFloat(BitPrec);
      vector<RingElem> semiwidth;
      vector<point> Pts;
      RingElem Gamma(zero(RR));
      ConversionToRR(OrigSemiwidth, OrigPts, OrigGamma, semiwidth, Pts, Gamma, RR);
      // Build poly rings P=RR[x_1...x_n] and 
      // PErr = RR[e[1,1],...,e[NumPts,dim]]
      SparsePolyRing P = NewPolyRing(RR, PPM);
      vector<int> First, Last;
      First.push_back(1); First.push_back(1);
      Last.push_back(NumPts); Last.push_back(dim);
      SparsePolyRing PErr = NewPolyRing(RR, SymbolRange(symbol("e",First), symbol("e",Last))); 
      // Build evaluation homomorphisms at Pts
      list<RingHom> EvalAtPts;
      for (size_t i=0; i < NumPts; ++i)
        EvalAtPts.push_back(EvalHom(P, Pts[i]));
	  
      //   --------------     INITIALIZE   ---------------
      size_t IterNum = 1;
      const RingElem Eps_square = squareNorm(semiwidth);
      const RingElem kappa = one(RR) + Gamma;
      const RingElem Bound = kappa*kappa*NumPts*Eps_square;
      vector<RingElem> J, S;     //J = almost vanishing polys
	  	  
      QBGenerator QBG(PPM);      //to handle quotient bases
      QBG.myCornerPPIntoQB(one(PPM));
      S.push_back(one(P));
      list<PPMonoidElem> L = QBG.myCorners();  //L = list of PP 
	  	  	     
      // Build and initialize matrices A0, A1 DiagA0, A0A1, CErr 
      matrix A0(NewDenseMat(RR, NumPts, NumPts));   //A0=orthogonal
      for (size_t i=0; i < NumPts; ++i)
        SetEntry(A0, 0, i, one(RR));

      matrix A1(NewDenseMat(PErr, NumPts, NumPts));

      vector<RingElem> DiagA0Elems(NumPts, zero(RR));
      MatrixView DiagA0(DiagMat(DiagA0Elems));//DiagA0=A0*A0'
      SetEntry(DiagA0, 0, 0, NumPts);

      vector<RingElem> A0A1Elems(NumPts, zero(PErr));
      MatrixView A0A1(DiagMat(A0A1Elems));//A0A1=2*A0*A1'


      // ------------   MAIN CYCLE   -----------------
      while (!L.empty())
      {
        GlobalLogput() << "---------------------------------------" 
                       << endl;
        GlobalLogput() << "Iteration " << IterNum << endl;
        GlobalLogput() << "Quotient Basis = " << QBG.myQB() << endl;
        GlobalLogput() << "List of power products = " << L << endl;
        const PPMonoidElem t = L.front(); //PP analyzed current step
        GlobalLogput() << "Next PP = " << t << endl;
        const RingElem tAsPoly = monomial(P, one(RR), t);

        //Compute eval V0=t(Pts) and solve LS problem A0'*b = V0
        vector<RingElem> V0,V1,Alpha0, Alpha1, Rho0, Rho1;
        Evaluation(V0, tAsPoly, EvalAtPts);
        LeastSquares(Alpha0, Rho0, A0, DiagA0, V0);
        // First trivial check   
        if (IsZero(Rho0))
        {
          AddToCornerAndUpdate(t, QBG, Alpha0, S, J);
          ++IterNum;
          L = QBG.myCorners(); 
          continue;
        }		

        //Compute f.o.eval V1, solve (f.o.error anal.)LS problem
        FirstOrderEval(V1, tAsPoly, EvalAtPts, semiwidth, PErr);
        FirstOrderLS(Rho1, Alpha1, A0, A1, DiagA0, A0A1, V0, V1, Alpha0);  
        //Compute CErr such that Rho1 = CErr*e 
        matrix CErr(NewDenseMat(RR, NumPts, NumPts*dim));//CErr=Rho1(matrix)
        ErrorMatrix(CErr, Rho1);
        //Compute the strongly independent rows of CErr
        vector<size_t> rows;		       
        WellIndepRows(rows, Eps_square, CErr);	      
        GlobalLogput() << "Well Independent Rows =" << rows << endl;

        //if, for each i, ||CErr[i]|| < Eps ...
        if (rows.empty())
        {
          if (squareNorm(Rho0) > Bound*Eps_square)
          {
            UpdateMatrices(QBG, A0, A1, DiagA0, A0A1,Rho0,Rho1);
            AddToQBAndUpdate(t, QBG, Alpha0, S);
          }
          else
            AddToCornerAndUpdate(t, QBG, Alpha0, S, J);
        }
        else
        {
          //Compute the submatrix of CErr with rows in rows
          vector<size_t> cols;
          for (size_t j=0; j < NumCols(CErr); ++j)
            cols.push_back(j);		  
          ConstMatrixView CErr1 = submat(CErr, rows, cols);
          vector<RingElem> MinusRho01;
          for (size_t i=0; i < rows.size(); ++i)
            MinusRho01.push_back(-Rho0[rows[i]]);
          vector<RingElem> eLS1;
          //Compute the minimal 2-norm solution of the 
          //underdetermined system CErr1 * eLS1 = MinusRho01
          SolveFRUDS(eLS1, CErr1, MinusRho01);
          GlobalLogput() << "Square norm eLS1= " 
                         << squareNorm(eLS1) << endl;
          GlobalLogput() << "Bound = " << Bound << endl;

          //Check and update the variables
          if (squareNorm(eLS1) >= Bound)
          {
            UpdateMatrices(QBG, A0, A1, DiagA0, A0A1,Rho0,Rho1);
            AddToQBAndUpdate(t, QBG, Alpha0, S);
          }			
          else
            AddToCornerAndUpdate(t, QBG, Alpha0, S, J);	  
        }
        ++IterNum;
        L = QBG.myCorners();
      }

      GlobalLogput() << "==========================================" 
                     << endl << endl;
      GlobalLogput() << "All elements of the SOI have been found" 
                     << endl;
      GlobalLogput() << "Stable order ideal: " << QBG.myQB() << endl 
                     << endl;

      QB = QBG.myQB();
      //Computation of the Border Basis and printing
      BBasis.clear();
      AlmostVanishing = J;
      if ((QBG.myQB()).size() == NumPts) 
      {
        list<PPMonoidElem> Border;
	//        vector<RingElem> BBasis;
        ComputeBorder(Border, QBG);
        for (list<PPMonoidElem>::const_iterator it=Border.begin(); it!=Border.end(); ++it)
        {
          RingElem f(P);
          ComputeBorderPoly(f, *it, A0, DiagA0, EvalAtPts, S);
          BBasis.push_back(f);
        }
	GlobalLogput() << "BBasis = " << BBasis << endl << endl;
        if (IsBorderBasis(BBasis, EvalAtPts))
        {
          GlobalLogput() << "BBasis successfully computed with "
                         << BitPrec << " bit precision " << endl;
	  GlobalLogput() << "Border Basis = " << BBasis << endl;
        }
        else
          GlobalLogput() << "Error: it is not a border basis" 
                         << endl;
 
        GlobalLogput() << endl << "Almost vanishing polys = " 
                       << J << endl;
      }
      else
      {
        GlobalLogput() << "Found " << (QBG.myQB()).size() 
                       << " elements of QB" << endl;
        GlobalLogput() << "A 'smaller' semiwidth should be used!" 
                       << endl;
        GlobalLogput() << endl << "Almost vanishing polys = " 
                       << J << endl;
      }
      break;   
    }
    catch (RingTwinFloat::InsufficientPrecision&)
    {
      // Inform the user about the failure...
      GlobalLogput() << endl << "----------------------------------" 
                     << endl << "A bit precision of " << BitPrec 
                     << " was not sufficient." << endl 
                     << "------------------------------------------" 
                     << endl << endl;
    }
  }
}


  } // end of namespace ApproxPts

} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/ApproxPts.C,v 1.6 2008/06/04 18:27:37 abbott Exp $
// $Log: ApproxPts.C,v $
// Revision 1.6  2008/06/04 18:27:37  abbott
// Modified the server interface for "SOI": it now accepts a 3rd arg (gamma).
//
// Revision 1.5  2008/05/30 14:20:43  abbott
// SOI now returns also the "almost vanishing" polynomials.
//
// Revision 1.4  2008/05/30 12:51:08  abbott
// Added a comment.
//
// Revision 1.3  2008/05/29 15:46:29  bigatti
// -- added Approximate Border Basis (by Abbott,Torrente)
//
// Revision 1.2  2007/10/30 17:14:08  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.8  2007/03/08 18:22:30  cocoa
// Just whitespace cleaning.
//
// Revision 1.7  2006/12/21 13:48:33  cocoa
// Made all increment/decrement calls prefix (except where the must be postfix).
//
// Revision 1.6  2006/11/22 14:43:32  cocoa
// -- minor cleaning (indicated by Intel compiler)
//
// Revision 1.5  2006/11/20 15:53:38  cocoa
// Minor cleaning.  Improved comments.
//
// Revision 1.4  2006/10/31 14:26:03  cocoa
// Added some missing consts; should make the code a little easier to comprehend.
//
// Revision 1.3  2006/10/06 14:04:15  cocoa
// Corrected position of #ifndef in header files.
// Separated CoCoA_ASSERT into assert.H from config.H;
// many minor consequential changes (have to #include assert.H).
// A little tidying of #include directives (esp. in Max's code).
//
// Revision 1.2  2006/06/21 17:05:47  cocoa
// Major overhaul of approx point preprocessing algms.
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.3  2006/05/29 16:16:47  cocoa
// Added "disj" preprocessing algorithm.
//
// Revision 1.2  2006/05/22 15:52:16  cocoa
// Added preprocess-disg algorithm to ApproxPts.
// Sundry minor improvements.
//
// Revision 1.1  2006/05/12 13:16:30  cocoa
// Added functions for preprocessing approximate points.
//
//
